/* 
 * impact_shadow.c 2005/07/12 23:24:15, Copyright (c) 2005 peter fuerst
 *
 * Based on:
 * - linux/drivers/video/impact.c, 2005 pf.
 * - linux/drivers/video/impactsr.c, (c) 2004 by Stanislaw Skowronek.
 * - newport_shadow.c
 *   # newport_shadow.c,v 1.3 2000/11/29 20:58:10 agx Exp #
 *   # xc/programs/Xserver/hw/xfree86/drivers/newport/newport_shadow.c,v
 *     1.2 2001/11/23 19:50:45 dawes Exp #
 */

#include "impact.h"


static __inline__ int
ImpactPreRefreshBox(mgicfifo_t *cfifo, const BoxPtr obox, BoxPtr nbox)
{
	int h, w, i;
	/*
	 * NB: Box defines [x1,x2[ x [x2,y2[, not [x1,x2] x [x2,y2] !
	 */
	*nbox = *obox;
	if (nbox->x1 < 0) nbox->x1 = 0;
	if (nbox->y1 < 0) nbox->y1 = 0;
	if (1280 < nbox->x2) nbox->x2 = 1280;
	if (1024 < nbox->y2) nbox->y2 = 1024;
	/*
	 * We may need to do some adjustment, since
	 * 1) the actual number of pixels *must* match the settings
	 *    in "xystarti", "xyendi", "xfrsize", "xfrcounters",
	 * 2) we draw pairs of pixels.
	 * Otherwise... (try it :))
	 */
	nbox->x1 &= ~1;
	nbox->x2 = ++nbox->x2 & ~1;
	w = nbox->x2 - nbox->x1;
	h = nbox->y2 - nbox->y1;
	if (w <= 0 || h <= 0)
		return 0;

	/* setup PIO to RE */
	impact_cmd_pp1fillmode(cfifo, 0x6300, IMPACT_LO_COPY);
	impact_cmd_blockxystarti(cfifo, nbox->x1, nbox->y1);
	impact_cmd_blockxyendi(cfifo, nbox->x2-1, nbox->y2-1);
	impact_cmd_fillmode(cfifo, 0x00c00000);
	impact_cmd_xfrmode(cfifo, 0x00080);
	impact_cmd_xfrsize(cfifo, w, h);
	impact_cmd_xfrcounters(cfifo, w, h);
	impact_cmd_gline_xstartf(cfifo, 1);
	impact_cmd_ir_alias(cfifo, 0x18);
	/* another workaround.. 33 writes to alpha... hmm... */
	for (i = 0; i < 33; i++)
		impact_cmd_alpha(cfifo, 0);
	impact_cmd_xfrcontrol(cfifo, 2);
	return 1;
}

static __inline__ void
ImpactPostRefreshBox(mgicfifo_t *cfifo)
{
	impact_cmd_gline_xstartf(cfifo, 0);
	impact_cmd_re_togglecntx(cfifo, 0);
	impact_cmd_xfrcounters(cfifo, 0, 0);
}

/*
 * 8-bpp blits are done as PIO draw operation; the pixels are unpacked
 * into 32-bpp values from the current palette in software.
 */
static void
ImpactRefreshArea8(ImpactPtr pImpact, mgicfifo_t *cfifo, int num, BoxPtr pbox)
{
	unsigned pal[256];
	int i;

	for (i = 0; i < sizeof(pal)/sizeof(*pal); i++)
		pal[i] = ImpactGetPalReg(pImpact, i);

	for (; num--; pbox++) {
		BoxRec sanebox;
		unsigned char *base;
		int dy;
		TRACEV("ImpactRefreshArea8: %d [%d..%d[x[%d..%d[\n",
				num, pbox->x1,pbox->x2, pbox->y1,pbox->y2 );

		(*pImpact->WaitCfifoEmpty)(pImpact->pImpactRegs);
		if (ImpactPreRefreshBox(cfifo,pbox,&sanebox) <= 0)
			continue;

		base = (unsigned char*)pImpact->ShadowPtr 
			+ (dy = sanebox.y1) * pImpact->ShadowPitch + sanebox.x1;

		for ( ; dy < sanebox.y2; dy++) {
			unsigned char *src = base;
			int dx = sanebox.x1;
			while (dx < sanebox.x2) {
				/* pairs of pixels are sent in two writes to the RE */
				impact_cmd_char_h(cfifo, pal[*src++]);
				impact_cmd_char_l(cfifo, pal[*src++]);
				dx += 2;
			}
			base += pImpact->ShadowPitch;
		}
		ImpactPostRefreshBox(cfifo);
	}
}

void
ImpactI2RefreshArea8(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
	if (num > 0)
	{	ImpactPtr pImpact = IMPACTPTR(pScrn);
		ImpactRegsPtr pImpactRegs = pImpact->pImpactRegs;

		ImpactRefreshArea8(pImpact, &pImpactRegs->i2.cfifo, num, pbox);
		TRACE_EXIT("ImpactRefreshArea8");
	}
}

void
ImpactSRRefreshArea8(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
	if (num > 0)
	{	ImpactPtr pImpact = IMPACTPTR(pScrn);
		ImpactRegsPtr pImpactRegs = pImpact->pImpactRegs;

		ImpactRefreshArea8(pImpact, &pImpactRegs->sr.cfifo, num, pbox);
		TRACE_EXIT("ImpactRefreshArea8");
	}
}

static __inline__ int
ImpactPreRefreshBoxDMA(mgicfifo_t *cfifo, const BoxPtr obox, BoxPtr nbox, unsigned bpitch)
{
	int h, w, i;
	/*
	 * NB: Box defines [x1,x2[ x [x2,y2[, not [x1,x2] x [x2,y2] !
	 */
	*nbox = *obox;
	if (nbox->x1 < 0) nbox->x1 = 0;
	if (nbox->y1 < 0) nbox->y1 = 0;
	if (1280 < nbox->x2) nbox->x2 = 1280;
	if (1024 < nbox->y2) nbox->y2 = 1024;

	nbox->x1 &= ~1;
	nbox->x2 = ++nbox->x2 & ~1;
	w = nbox->x2 - nbox->x1;
	h = nbox->y2 - nbox->y1;
	if (w <= 0 || h <= 0)
		return 0;

	/* setup DMA to RE */
	impact_cmd_pp1fillmode(cfifo, 0x6300, IMPACT_LO_COPY);
	impact_cmd_colormasklsbsa(cfifo, 0xFFFFFFFF);
	impact_cmd_colormasklsbsb(cfifo, 0xFFFFFFFF);
	impact_cmd_colormaskmsbs(cfifo, 0xFFFF);
	impact_cmd_xfrmasklo(cfifo, 0xFFFFFFFF);
	impact_cmd_xfrmaskhi(cfifo, 0xFFFFFFFF);
	impact_cmd_packedcolor(cfifo, 0);
	impact_cmd_red(cfifo, 0);
	impact_cmd_drbpointers(cfifo, 0xc8240);
	impact_cmd_blockxystarti(cfifo, nbox->x1, nbox->y1);
	impact_cmd_blockxyendi(cfifo, nbox->x2-1, nbox->y2-1);
	impact_cmd_fillmode(cfifo, 0x01400000);
	impact_cmd_xfrmode(cfifo, 0x00080);
	impact_cmd_hq_pixelformat(cfifo, 0x600);
	impact_cmd_hq_scanwidth(cfifo, w<<2);
	impact_cmd_hq_dmatype(cfifo, 0x0c);
	impact_cmd_pixcmd(cfifo, 3);
	impact_cmd_xfrsize(cfifo, w, h);
	impact_cmd_xfrcounters(cfifo, w, h);
	impact_cmd_gline_xstartf(cfifo, 1);
	impact_cmd_ir_alias(cfifo, 0x18);
	for (i = 0; i < 33; i++)
		impact_cmd_alpha(cfifo, 0);
	impact_cmd_xfrcontrol(cfifo, 1);
	impact_cmd_hq_pg_list0(cfifo, 0x80000000);
	impact_cmd_hq_pg_width(cfifo, bpitch);
	impact_cmd_hq_pg_offset(cfifo, 0);
	impact_cmd_hq_pg_startaddr(cfifo, bpitch*nbox->y1+(nbox->x1<<2));
	impact_cmd_hq_pg_linecnt(cfifo, h);
	impact_cmd_hq_pg_widtha(cfifo, w<<2);
	impact_cmd_hq_dmactrl_1(cfifo);
	impact_cmd_hq_dmactrl_2(cfifo);

	return 1;
}

static __inline__ void
ImpactPostRefreshBoxDMA(mgicfifo_t *cfifo)
{
	impact_cmd_gline_xstartf(cfifo, 0);
	impact_cmd_re_togglecntx(cfifo, 0);
	impact_cmd_xfrcounters(cfifo, 0, 0);
	impact_cmd_pixcmd(cfifo, 0);
	impact_cmd_hq_pixelformat(cfifo, 0xe00);
}

/* 32-bit blits are done as DMA operations, which is FAST on SGI machines */
static void
ImpactRefreshArea32(ImpactPtr pImpact, mgicfifo_t *cfifo, int num, BoxPtr pbox)
{
	for (; num--; pbox++) {
		BoxRec sanebox;
		TRACEV("ImpactRefreshArea32: %d [%d..%d]x[%d..%d]\n",
				num, pbox->x1,pbox->x2, pbox->y1,pbox->y2 );

		(*pImpact->WaitDMAOver)(pImpact->pImpactRegs);
		if (ImpactPreRefreshBoxDMA(cfifo,pbox,&sanebox, pImpact->ShadowPitch) <= 0)
			continue;
		(*pImpact->WaitDMAOver)(pImpact->pImpactRegs);
		ImpactPostRefreshBoxDMA(cfifo);
	}
}

void
ImpactI2RefreshArea32(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
	if (num > 0)
	{	ImpactPtr pImpact = IMPACTPTR(pScrn);
		ImpactRegsPtr pImpactRegs = pImpact->pImpactRegs;

		ImpactRefreshArea32(pImpact, &pImpactRegs->i2.cfifo, num, pbox);
		TRACE_EXIT("ImpactRefreshArea32");
	}
}

void
ImpactSRRefreshArea32(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
	if (num > 0)
	{	ImpactPtr pImpact = IMPACTPTR(pScrn);
		ImpactRegsPtr pImpactRegs = pImpact->pImpactRegs;

		ImpactRefreshArea32(pImpact, &pImpactRegs->sr.cfifo, num, pbox);
		TRACE_EXIT("ImpactRefreshArea32");
	}
}

/* eof */
